from typing import List
import logging
import sys
from pathlib import Path
from pprint import pformat

import click
from rich.console import Console

from f_table import get_table, BasicScreenStyle
from git_wrap import GitRepo, get_git_dirs

from au.classroom import Assignment, Roster, get_accepted_assignments
from au.click import BasePath, AssignmentOptions, RosterOptions, DebugOptions
from au.common import draw_double_line, draw_single_line


logger = logging.getLogger(__name__)


@click.command("clone-all")
@click.argument("root_dir", type=BasePath(), default=".")
@AssignmentOptions(required=True).options
@RosterOptions(prompt=True).options
@click.option(
    "--preserve-prefix",
    is_flag=True,
    help="set to preserve the prefix (slug) string common to all repositories",
)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="set to pull changes to existing repositories",
)
@click.option(
    "-y",
    "--skip-confirm",
    is_flag=True,
    help="set to bypass confirmation and overwrite existing CSV",
)
@click.option(
    "-p",
    "--preview",
    is_flag=True,
    help="set to show changes without actually making them",
)
@DebugOptions().options
def clone_all_cmd(
    root_dir: Path,
    assignment: Assignment = None,
    roster: Roster = None,
    preserve_prefix: bool = False,
    update: bool = False,
    skip_confirm: bool = False,
    preview: bool = False,
    **kwargs,
):
    """Clone all student repos for an assignment into ROOT_DIR.

    This command will use the specified assignment to query GitHub Classroom for
    all accepted assignments. For each, it will run the following logic:

    \b
        if repository directory exists in ROOT_DIR
            if update flag is set
                pull updates from GitHub
            else
                skip
        else
            if repository has submissions
                clone repository into ROOT_DIR
            else
                skip

    If no roster is provided, directory names will match the students' GitHub
    login names. If a GitHub Classroom roster is provided using the --roster
    option and a student name match can be found in the roster, then the
    directory will be named as follows:

        [student_name]@[login_name]

    All commas, spaces, and invalid characters in the student name will be
    replaced with underscores. for example:

        York_Paul@ptyork

    If the `--preserve-prefix` flag is set, then the prefix (also called 'slug')
    will be added to all assignments, as they are in GitHub. This will insure
    that all student assignment directory names are unique.

    If ROOT_DIR is not provided, then the current working directory will be
    assumed.
    """
    logging.basicConfig()

    draw_double_line()
    print(assignment)
    draw_single_line()
    if roster:
        print(f"USING ROSTER: {roster.file.resolve()}")
    else:
        print("NOT USING ROSTER")
    draw_single_line()

    if not skip_confirm:
        if not click.confirm(
            f"Are you sure you want to clone the above assignment into {root_dir}?"
        ):
            sys.exit(0)

    clone_all(root_dir, assignment, roster, preserve_prefix, update, preview)


def clone_all(
    root_dir: Path,
    assignment: Assignment,
    roster: Roster = None,
    preserve_prefix: bool = False,
    update: bool = False,
    preview: bool = False,
):
    """clone all student submissions for an assignment into root_dir."""
    if preview:
        print()
        draw_double_line()
        draw_single_line("PREVIEW ONLY - NO CHANGES BEING MADE")
        draw_double_line()
        print()

    if assignment.type not in ["individual", "group"]:
        logger.fatal("Only 'individual' and 'group' assignments are supported")
        sys.exit(1)

    assignment_prefix = assignment.slug + "-"

    if roster and assignment.type != "individual":
        logger.warning(
            "Use of GitHub classroom roster only valid for individual assignments."
        )
        roster = None

    console = Console()
    with console.status(
        "Retrieving data from GitHub Classroom", spinner="bouncingBall"
    ) as status:
        accepted_assignments = get_accepted_assignments(assignment)

        submitted_assignments = [a for a in accepted_assignments if a.commit_count > 0]
        unsubmitted_assignments = [
            a for a in accepted_assignments if a.commit_count == 0
        ]

        if assignment.type == "individual":
            # Get the student name for individual assignments
            accepted_logins = [a.students[0].login for a in accepted_assignments]
            submitted_logins = [a.students[0].login for a in submitted_assignments]
            unsubmitted_logins = [a.students[0].login for a in unsubmitted_assignments]
            login_url_map = {
                a.students[0].login: a.repository.html_url
                for a in submitted_assignments
            }
        else:
            # Get the group name for group assignments
            accepted_logins = [
                a.repository.name.removeprefix(assignment_prefix)
                for a in accepted_assignments
            ]
            submitted_logins = [
                a.repository.name.removeprefix(assignment_prefix)
                for a in submitted_assignments
            ]
            unsubmitted_logins = [
                a.repository.name.removeprefix(assignment_prefix)
                for a in unsubmitted_assignments
            ]
            login_url_map = {
                a.repository.name.removeprefix(assignment_prefix): a.repository.html_url
                for a in submitted_assignments
            }
        if not roster:
            roster = Roster(accepted_logins)
        else:
            # Add all...will ignore ones already in the roster
            for login in accepted_logins:
                roster.append(login)

        status.update(status="Checking status of existing repositories")

        existing_repo_dirs = get_git_dirs(root_dir)

        login_all_dir_map = roster.get_login_dir_map(root_dir)
        login_pull_dir_map = roster.get_login_dir_map(existing_repo_dirs)
        login_bad_dir_map = {
            login: path
            for login, path in login_all_dir_map.items()
            if login not in login_pull_dir_map
        }

        login_clone_dir_map = roster.get_dir_names(
            assignment_prefix if preserve_prefix else None
        )
        for login in roster.get_logins():
            if login not in submitted_logins:
                login_clone_dir_map.pop(login, None)
        for login in login_pull_dir_map:
            login_clone_dir_map.pop(login, None)
        for login in login_bad_dir_map:
            login_clone_dir_map.pop(login, None)

        clones: List[str] = []
        pulls: List[str] = []
        skips: List[str] = []
        errors: List[str] = []

        logger.debug("login_clone_dir_map: " + pformat(login_clone_dir_map))
        logger.debug("login_pull_dir_map" + pformat(login_all_dir_map))
        logger.debug("login_bad_dir_map" + pformat(login_bad_dir_map))

        for login, dir_name in login_clone_dir_map.items():
            repo_url = login_url_map[login]
            status.update(status=f"CLONING {dir_name} from {repo_url}")
            if not preview:
                repo_path = root_dir / dir_name
                try:
                    GitRepo.clone(repo_url, repo_path)
                except:
                    logger.exception(
                        f"Exception raised while cloning from {repo_url} into {repo_path}"
                    )
                    errors.append(roster.get_name(login))

        if update:
            for login, dir_name in login_pull_dir_map.items():
                try:
                    repo = GitRepo(root_dir / dir_name)
                    if preview:
                        status.update(status=f"Checking remote status for {dir_name}")
                        if repo.needs_pull():
                            pulls.append(roster.get_name(login))
                        else:
                            skips.append(roster.get_name(login))
                    else:
                        status.update(status=f"PULLING {dir_name}")
                        if repo.pull():
                            pulls.append(roster.get_name(login))
                        else:
                            skips.append(roster.get_name(login))
                except:
                    logger.exception(f"Exception raised while pulling {dir_name}")
                    errors.append(roster.get_name(login))
        else:
            pulls = [roster.get_name(l) for l in login_pull_dir_map]

    unsubmitted = [roster.get_name(login) for login in unsubmitted_logins]
    unaccepted = [
        roster.get_name(login)
        for login in roster.get_logins()
        if login not in accepted_logins
    ]

    summary_rows = []
    if clones:
        clones.sort()
        summary_rows.append([f"Cloned ({len(clones)})", "\n".join(clones)])
    else:
        summary_rows.append(
            ["Pulled (0)", "No uncloned repositories for this assignment"]
        )
    if pulls:
        pulls.sort()
        pull_str = "Pulled" if update else "Not Pulled: no --update"
        summary_rows.append([f"{pull_str} ({len(pulls)})", "\n".join(pulls)])
    if skips:
        skips.sort()
        summary_rows.append(
            [f"Not Pulled - No Changes ({len(skips)})", "\n".join(skips)]
        )
    if unsubmitted:
        unsubmitted.sort()
        summary_rows.append(
            [f"Not Submitted ({len(unsubmitted)})", "\n".join(unsubmitted)]
        )
    if unaccepted:
        unaccepted.sort()
        summary_rows.append(
            [f"Not Accepted ({len(unaccepted)})", "\n".join(unaccepted)]
        )
    if errors:
        errors.sort()
        summary_rows.append([f"Errors ({len(errors)})", "\n".join(errors)])

    print(get_table(summary_rows, style=BasicScreenStyle(), separete_rows=True))


if __name__ == "__main__":
    clone_all_cmd()



from typing import Dict, List
from pathlib import Path
import logging
import re

from au.common import dict_from_csv


logger = logging.getLogger(__name__)

_invalid_name_char_pattern = re.compile(r'([<>:"\/\\\,|?*]|\s)+')


class Roster:
    def __init__(self, source: Path | List):
        if isinstance(source, Path):
            if source and source.exists():
                source = source.resolve()
                self.login_student_map: Dict[str, str] = dict_from_csv(
                    source, "github_username", "identifier"
                )
                for login in self.login_student_map:
                    if not self.login_student_map[login]:
                        self.login_student_map[login] = login
                self.file: Path = source
            else:
                raise FileNotFoundError(f"Roster file {source} not found.")
        elif isinstance(source, list):
            # source will be a list of logins or repository names, so just make the name and login the same
            self.login_student_map: Dict[str, str] = {str(v): str(v) for v in source}
        else:
            raise ValueError(str(source), "is not a valid roster source")

    def get_logins(self) -> List[str]:
        return list(self.login_student_map.keys())

    def get_names(self) -> List[str]:
        return list(self.login_student_map.values())

    def get_name(self, login: str) -> str:
        return self.login_student_map.get(login)

    def get_login_student_map(self):
        return self.login_student_map.copy()

    def get_student_login_map(self):
        student_login_map = {v: k for k, v in self.login_student_map.items()}
        if not len(self.login_student_map) == len(student_login_map):
            logger.warning("Duplicate student names found.")
        return student_login_map

    def append(self, login: str, name: str = None):
        """Add a login to the roster."""
        if login not in self.login_student_map:
            student = name if name else login
            self.login_student_map[login] = student

    def get_dir_names(self, prefix: str = None) -> Dict[str, str]:
        """Get friendly directory names for each student."""
        login_dir_name_map: Dict[str, str] = {}
        prefix = prefix if prefix else ""
        for login, student_name in self.login_student_map.items():
            if student_name != login:
                student_name = _invalid_name_char_pattern.sub("_", student_name)
                student_name = student_name.replace("__", "_")
                login_dir_name_map[login] = prefix + student_name + "@" + login
            else:
                login_dir_name_map[login] = prefix + login
        return login_dir_name_map

    @staticmethod
    def _get_dirs_from_source(
        dir_source: Path | List[any], sort_by_size=False
    ) -> List[str]:
        if isinstance(dir_source, Path):
            dirs: list[str] = []
            for subdir in dir_source.iterdir():
                # Only dirs
                if not subdir.is_dir():
                    continue
                # No hidden or special dirs
                if subdir.name[0] in "_.":
                    continue
                dirs.append(subdir.name)
        else:
            try:
                dirs = [d.name for d in dir_source]
            except:
                dirs = [str(d) for d in dir_source]

        if sort_by_size:
            dirs.sort(key=lambda l: len(l), reverse=True)
        else:
            dirs.sort()
        return dirs

    def get_login_dir_map(
        self, dir_source: Path | List[str], include_unmapped=False
    ) -> Dict[str, str | None]:
        """
        Map student logins to subdirectories or root_dir.

        In the rare case of multiple matches, only the first matching
        subdirectory is mapped.
        """
        logins = list(self.login_student_map.keys())
        logins.sort(key=lambda l: len(l), reverse=True)
        login_dir_map: Dict[str, str | None] = {}
        if include_unmapped:
            login_dir_map = {k: None for k in logins}
        subdirs = Roster._get_dirs_from_source(dir_source)
        for subdir in subdirs:
            for login in logins:
                if login in subdir and not login_dir_map.get(login):
                    login_dir_map[login] = subdir
                    break
        return login_dir_map

    def get_dir_login_map(self, dir_source: Path | List[str]) -> Dict[str, str | None]:
        """
        Map subdirectories of root_dir to student logins.

        In the rare case of multiple matches, all matching subdirectories
        will be mapped to a matching login.
        """
        subdirs = Roster._get_dirs_from_source(dir_source)
        dir_login_map: Dict[str, str | None] = {}
        logins = list(self.login_student_map.keys())
        logins.sort(key=lambda l: len(l), reverse=True)
        for subdir in subdirs:
            dir_login_map[subdir] = None
            for login in logins:
                if login in subdir:
                    dir_login_map[subdir] = login
        return dir_login_map

    def get_dir_student_map(
        self, dir_source: Path | List[str]
    ) -> Dict[str, str | None]:
        """
        Just a convenience method to map dirs directly to student names instead
        of logins.
        """
        dir_login_map = self.get_dir_login_map(dir_source)
        dir_student_map: Dict[str, str | None] = {}
        for dirname, login in dir_login_map.items():
            student = None
            if login:
                student = self.login_student_map[login]
            dir_student_map[dirname] = student
        return dir_student_map

    def get_login_for_dir(self, path: Path) -> str:
        dir_name = path.resolve().name
        logins = list(self.login_student_map.keys())
        logins.sort(key=lambda l: len(l), reverse=True)
        for login in logins:
            if login in dir_name:
                return login
        return None



from typing import Dict
import logging
import re
from pathlib import Path
from textwrap import wrap, indent
from pprint import pformat

import click

from au.click import BasePath, DebugOptions

from .eval_assignment import retrieve_student_results
from .pytest_data import Results, Test, SubTest
from .scoring import get_summary, get_summary_row, get_student_scores, ScoringParams


logger = logging.getLogger(__name__)


DEFAULT_FEEDBACK_FILE_NAME = "FEEDBACK.md"


def get_feedback_file_score(feedback_file_path: Path) -> str:
    """
    Returns score as string in case it has been annotated in the feedback file.
    """
    score = None
    with open(feedback_file_path, "r") as fi:
        for line in fi:
            if line.startswith("| Final Score"):
                parts = line[2:].split("|")
                score = parts[1].strip()
                break
    return score


@click.command("gen-feedback")
@click.argument("student_dir", type=BasePath(), required=True)
@click.option(
    "--feedback-filename",
    type=str,
    default=DEFAULT_FEEDBACK_FILE_NAME,
    help="name of markdown file to generate",
)
@click.option(
    "-o",
    "--overwrite-feedback",
    is_flag=True,
    help="set to override default behavior of not overwriting feedback files",
)
@click.option(
    "-max",
    "--max-score",
    type=float,
    default=10,
    show_default=True,
    help="the maximum score for this assignment",
)
@click.option(
    "-ptw",
    "--pytest-weight",
    type=float,
    default=1.0,
    show_default=True,
    help="the weight to apply to pytest when calculating the overall score (0 to 1)",
)
@click.option(
    "-plw",
    "--pylint-weight",
    type=float,
    default=0.0,
    show_default=True,
    help="the weight to apply to pylint when calculating the overall score (0 to 1)",
)
@DebugOptions().options
def gen_feedback_cmd(
    student_dir: Path,
    feedback_filename: str = DEFAULT_FEEDBACK_FILE_NAME,
    overwrite_feedback: bool = False,
    max_score: int = 10,
    pytest_weight: float = 1.0,
    pylint_weight: float = 0.0,
    **kwargs,
) -> None:
    """Generate a feedback file for a single assignment."""
    logging.basicConfig()

    try:
        student_results = retrieve_student_results(student_dir)
    except:
        logger.error(f"Unable to find results file. Have you run the tests yet?")

    #################################################################
    # TODO: Pull feedback filename and scoring params from settings

    scoring_params = ScoringParams(max_score, pytest_weight, pylint_weight)
    gen_feedback(
        student_results,
        student_dir,
        feedback_filename,
        scoring_params,
        overwrite_feedback,
    )


def gen_feedback(
    student_results: Dict[str, any],
    student_dir: Path,
    feedback_filename: str = DEFAULT_FEEDBACK_FILE_NAME,
    scoring_params: ScoringParams = ScoringParams(),
    overwrite_feedback=False,
) -> None:
    """Generate a feedback file for a single student."""
    feedback_file_path = (student_dir / feedback_filename).resolve()

    if feedback_file_path.exists() and not overwrite_feedback:
        logger.info(
            f"SKIPPING: {feedback_file_path} already exists and --overwrite-feedback not specified"
        )
        return

    summary = get_summary(student_results, scoring_params, get_markdown=True)
    scores = get_student_scores(student_results, scoring_params)
    final_score_row = get_summary_row("Final Score", f"{scores.overall_score:g}")

    with open(feedback_file_path, "w") as fi:

        def wl(txt=""):
            fi.write(str(txt) + "\n")

        regex_clean = re.compile(r".+Error:.+?\s:", re.DOTALL)

        def print_test(test: Test | SubTest):
            if not test.is_passing():
                full_name = test.parent_test_class
                if isinstance(test, SubTest):
                    full_name += " >> " + test.parent_test_name
                full_name += " >> " + (test.name if test.name else "Unnamed Test")
                message = regex_clean.sub("", test.message).strip()
                wl(f"{full_name}")
                wl(indent(message, "    "))
                wl()
                wl(" •" * 20)
                wl()

        wl("# Assignment Feedback")
        wl()
        wl(summary)
        wl(final_score_row)
        wl()
        wl()
        wl("-" * 80)
        wl("## Grader Comments:")
        wl()
        wl()
        wl()

        pytest_results_d = student_results.get("pytest_results")
        if pytest_results_d:
            logger.debug("pytest_results: " + pformat(pytest_results_d))
            pytest_results = Results.from_dict(pytest_results_d)
            if pytest_results.pass_pct < 1:
                wl()
                wl("-" * 80)
                wl("## Functionality Feedback (pytest)")
                wl("-" * 80)
                wl("```")
                for test_class in pytest_results.test_classes.values():
                    if test_class.is_passing():
                        continue
                    for test in test_class.tests.values():
                        if test.is_passing():
                            continue
                        if test.sub_tests:
                            for sub_test in test.sub_tests:
                                if sub_test.get("status") == "pass":
                                    continue
                                print_test(sub_test)
                        else:
                            print_test(test)
            wl("```")

        pylint_results = student_results.get("pylint_results")
        if pylint_results:
            logger.debug("pylint_results: " + pformat(pylint_results))
            wl()
            wl("-" * 80)
            wl("## Code Style Feedback (pylint)")
            wl()
            wl("```")
            messages = pylint_results.get("messages", [])
            msg: Dict
            for msg in messages:
                path = msg.get("path", "General Message")
                line = msg.get("line", 0)
                line_str = f" line: {line}" if line else ""
                col = msg.get("col", 0)
                col_str = f" col: {col}" if col else ""
                id = msg.get("messageId", "No ID")
                msg_type = msg.get("type", "")
                message = msg.get("message", "No message provided")
                wl(f"{msg_type}: {path}{line_str}{col_str} ({id}):")
                wl(
                    "\n".join(
                        wrap(
                            message, 80, initial_indent="    ", subsequent_indent="    "
                        )
                    )
                )
                wl()
                wl(" •" * 20)
                wl()

            # wl(pformat(student_results['pylint_results']))
            wl("```")


if __name__ == "__main__":
    gen_feedback_cmd()



import logging
from pathlib import Path
from pprint import pformat

import click
from rich.console import Console

from git_wrap import get_git_dirs

from au.click import BasePath, AssignmentOptions, RosterOptions, DebugOptions
from au.classroom import Assignment, Roster
from au.common import draw_double_line, draw_single_line

from .eval_assignment import retrieve_student_results, eval_assignment
from .gen_feedback import (
    gen_feedback,
    get_summary,
    ScoringParams,
    DEFAULT_FEEDBACK_FILE_NAME,
)


logger = logging.getLogger(__name__)


@click.command()
@click.argument("root_dir", type=BasePath(), default=".")
@AssignmentOptions().options
@RosterOptions(prompt=True).options
@click.option(
    "-se", "--skip_eval", is_flag=True, help="set to bypass running the evaluations"
)
@click.option(
    "-sf", "--skip_feedback", is_flag=True, help="set to bypass generating feedback"
)
@click.option(
    "--feedback-filename",
    type=str,
    default=DEFAULT_FEEDBACK_FILE_NAME,
    show_default=True,
    help="name of markdown file to generate",
)
@click.option(
    "-o",
    "--overwrite-feedback",
    is_flag=True,
    help="set to override default behavior of not overwriting feedback files",
)
@click.option(
    "-max",
    "--max-score",
    type=float,
    default=10,
    show_default=True,
    help="the maximum score for this assignment",
)
@click.option(
    "-ptw",
    "--pytest-weight",
    type=float,
    default=1.0,
    show_default=True,
    help="the weight to apply to pytest when calculating the overall score (0 to 1)",
)
@click.option(
    "-plw",
    "--pylint-weight",
    type=float,
    default=0.0,
    show_default=True,
    help="the weight to apply to pylint when calculating the overall score (0 to 1)",
)
@DebugOptions().options
def quick_grade(
    root_dir: Path,
    assignment: Assignment = None,
    roster: Roster = None,
    skip_eval: bool = False,
    skip_feedback: bool = False,
    feedback_filename: str = DEFAULT_FEEDBACK_FILE_NAME,
    overwrite_feedback: bool = False,
    max_score: int = 10,
    pytest_weight: float = 1.0,
    pylint_weight: float = 0.0,
    **kwargs,
) -> None:
    """Run tests and generate feedback for all subdirectories of ROOT_DIR.

    This is essentially the same as:

    \b
        for SUBDIR in ROOT_DIR
            au python eval-assignment SUBDIR
            au python gen-feedback SUBDIR

    If ROOT_DIR is not provided, then the current working directory will be
    assumed.
    """
    logging.basicConfig()

    draw_double_line()
    if assignment:
        print(assignment)
    else:
        logger.error(
            "Unable to find the requested assignment. Functionality will be limited."
        )
    draw_single_line()

    dir_student_map = None
    if roster:
        print(f"USING ROSTER: {roster.file.resolve()}")
        login_student_map = roster.login_student_map
        logger.debug(pformat(login_student_map))
        dir_student_map = roster.get_dir_student_map(root_dir)
        logger.debug(pformat(dir_student_map))
    else:
        print("NOT USING ROSTER")

    draw_single_line()
    print()

    #################################################################
    # TODO: Pull feedback filename and scoring params from settings

    scoring_params = ScoringParams(max_score, pytest_weight, pylint_weight)

    ###############################################################################
    # PROCESS DIRS
    ###############################################################################

    console = Console()
    with console.status(
        status="Finding assignment directories in {root_dir}", spinner="bouncingBall"
    ):
        student_repos = get_git_dirs(root_dir)
        student_repos.sort()

    print(f"Processing {len(student_repos)} assignment directories")

    for student_repo in student_repos:

        print()
        dir_name = student_repo.name
        draw_double_line(f"Processing {dir_name}")

        student_name = None
        if dir_student_map and dir_name in dir_student_map:
            student_name = dir_student_map[dir_name]

        dir_path = (root_dir / student_repo).resolve()
        if skip_eval:
            try:
                student_results = retrieve_student_results(dir_path)
            except:
                print(f"SKIPPING: No results file found. Have you run the tests yet?")
                continue
        else:
            student_results = eval_assignment(dir_path, student_name, assignment)

        if not student_results:
            continue

        if not skip_feedback:
            draw_single_line(f"Generating {feedback_filename}")

            try:
                scoring_params = ScoringParams(max_score, pytest_weight, pylint_weight)
                gen_feedback(
                    student_results,
                    student_repo,
                    feedback_filename,
                    scoring_params,
                    overwrite_feedback,
                )
            except:
                logging.exception(
                    "An unexpected error occurred generating {student_repo / feedback_filename}"
                )

                print(f"done generating {feedback_filename}")

        print(get_summary(student_results, scoring_params))
        draw_double_line()
        print()


if __name__ == "__main__":
    quick_grade()



from .f_table import (
    get_table,
    get_table_row,
    get_table_header,
    ColDef,
    ColDefList,
    InvalidTableError,
    InvalidColDefError,
)

from .table_styles import (
    BoxChars,
    TableStyle,
    BasicScreenStyle,
    RoundedBorderScreenStyle,
    NoBorderScreenStyle,
    MarkdownStyle,
)



from typing import Dict, List
from dataclasses import dataclass, field
import logging
import sys
from pathlib import Path
from pprint import pformat

import click
from rich.console import Console

from f_table import get_table, BasicScreenStyle
from git_wrap import GitRepo

from au.classroom import Assignment, get_accepted_assignments, Roster
from au.click import BasePath, AssignmentOptions, RosterOptions, DebugOptions
from au.common import draw_double_line
from au.common.datetime import get_friendly_local_datetime, get_friendly_timedelta


logger = logging.getLogger(__name__)


@click.command()
@click.argument("root_dir", type=BasePath(), default=".")
@AssignmentOptions().options
@RosterOptions(prompt=True).options
@click.option("--late-only", is_flag=True, help="set to only show late students")
@DebugOptions().options
def time_details(
    root_dir: Path,
    assignment: Assignment = None,
    roster: Roster = None,
    late_only: bool = False,
    **kwargs,
):
    """Show submission times for the assignment in ROOT_DIR.

    Details include how long the assignment took to complete and how late (if at
    all) an assignment was.

    If ROOT_DIR is not provided, then the current working directory will be
    assumed.
    """
    logging.basicConfig()

    if not assignment:
        logger.fatal("Unable to find the requested assignment.")
        sys.exit(1)

    draw_double_line()
    print(assignment)

    console = Console()
    with console.status(
        status="Retrieving data from GitHub Classroom", spinner="bouncingBall"
    ) as status:
        # Pull logins to map to the roster, but only if the roster is provided and
        # this is an individual assignment
        accepted_logins = None
        dir_login_map = None
        if roster and assignment.type == "individual":
            # Attempt to get logins for all students to match against the dirs
            accepted = get_accepted_assignments(assignment)
            accepted_logins = [a.students[0].login for a in accepted]
            logger.debug("accepted_logins: " + pformat(accepted_logins))

            dir_login_map = roster.get_dir_login_map(root_dir)

        status.update(status="Finding all local Git repositories")

        repo_dirs: Dict[str, GitRepo] = {}
        for sub_dir in root_dir.iterdir():
            if not sub_dir.is_dir():
                continue
            if sub_dir.name.startswith(("_", ".")):
                continue
            try:
                repo_dirs[sub_dir.name] = GitRepo(sub_dir)
            except:
                continue

        logger.debug("repo_dirs: " + pformat(repo_dirs.keys()))

        @dataclass
        class _Commit:
            date: str
            message: str
            late: bool

        @dataclass
        class _Submission:
            name: str
            date: str = ""
            work_time: str = ""
            past_due: str = ""
            commits: List[_Commit] = field(default_factory=list)

            def add_commit(self, date, message, late):
                self.commits.append(_Commit(date, message, late))

        submissions: List[_Submission] = []
        self_email = GitRepo.get_user_email()

        status.update(status="Gathering time details from each repository")

        for dir_name, repo in repo_dirs.items():
            submission = _Submission(dir_name)
            submissions.append(submission)
            if dir_login_map:
                login = dir_login_map.get(dir_name)
                if login:
                    submission.name = roster.get_name(login)

            last_student_commit_date = None  # assume user submission
            for commit in repo.get_commits():
                if commit.author_email == self_email:
                    continue
                if "GitHub" == commit.committer_name:
                    if last_student_commit_date:
                        submission.work_time = get_friendly_timedelta(
                            last_student_commit_date - commit.date
                        )
                    break
                # assume a student commit
                date_str = get_friendly_local_datetime(commit.date)
                message = commit.message.strip()
                late = commit.date > assignment.deadline
                submission.add_commit(date_str, message, late)
                if not last_student_commit_date:
                    last_student_commit_date = commit.date
                    submission.date = date_str
                    if late:
                        submission.past_due = get_friendly_timedelta(
                            last_student_commit_date - assignment.deadline
                        )

    submissions.sort(key=lambda s: s.name)

    # FINALLY print the results
    rows = []
    for submission in submissions:
        if late_only and (not submission.past_due and submission.commits):
            continue
        past_due = "• • • • • • •"
        time_to_complete = "• • • • • • •"
        date_str = "• • • • • • • • • •"
        message = "• • • • • • • • • •"
        if submission.commits:
            past_due = submission.past_due
            time_to_complete = submission.work_time
            date_str = submission.commits[0].date
            message = submission.commits[0].message
        rows.append([submission.name, past_due, time_to_complete, date_str, message])
        # if late, also show the prior 4 commits
        if submission.past_due:
            for commit in submission.commits[1:5]:
                rows.append(["", "", "", commit.date, commit.message])

    print(
        get_table(
            rows,
            header_row=[
                "NAME",
                "PAST DUE",
                "WORK TIME",
                "COMMIT DATE",
                "COMMIT MESSAGE",
            ],
            col_defs=["", "", "", "", "AT"],
            style=BasicScreenStyle(),
        )
    )


if __name__ == "__main__":
    time_details()



from typing import List, Dict
from dataclasses import dataclass, field, asdict
from enum import Enum, auto
import json
import re
from dacite import from_dict, Config

class Status(Enum):
    """The status of a given test or test session."""
    PASS = "pass"
    FAIL = "fail"
    ERROR = "error"

@dataclass
class NodeId:
    test_file: str
    test_class: str
    test_name: str

    @staticmethod
    def parse(nodeid: str) -> 'NodeId':
        # TEST_FILE::TEST_CLASS::TEST_METHOD
        parts = nodeid.split('::')
        return NodeId(parts[0], parts[1], parts[2].replace('_', ' ').title())

_head_line_re = re.compile(r'^(\w*)\.(\w*)(?:\s+\[(.*)\])?(?:\s+\((.*)\))?$')

@dataclass
class SubTest:
    name: str
    parent_test_name: str
    parent_test_class: str
    status: Status = Status.PASS
    message: str|None = None
    output: str|None = None
    duration: float = 0.0

    def fail(self, message: str = None) -> None:
        """Indicate this test failed."""
        self.status = Status.FAIL
        self.message = message

    def error(self, message: str = None) -> None:
        """Indicate this test encountered an error."""
        self.status = Status.ERROR
        self.message = message

    def is_passing(self):
        """Check if the test is currently passing."""
        return self.status is Status.PASS


    @staticmethod
    def parse_head_line(head_line: str) -> str:
        """
        Returns either a subtest name or None depending on whether the head_line
        shows this to be a subtest or not. If the name is empty, "Unnamed
        Subtest" will be returned. There is no guarantee that subtests will have
        unique names.

        format:
            `TEST_CLASS.TEST_METHOD [SUBTEST_NAME] (SUBTEST_ARGS)`
        """
        parts = _head_line_re.match(head_line).groups()
        if parts[2] is None:
            return None
        else:
            name = parts[2].strip()
            if not name:
                name = "Unnamed Subtest"
            if parts[3]:
                name = f"{name} ({parts[3]})"
            return name


@dataclass
class Test:
    """An individual test's results."""

    name: str
    parent_test_class: str
    status: Status = Status.PASS
    message: str|None = None
    output: str|None = None
    duration: float = 0.0
    sub_tests: List[SubTest] = field(default_factory=list)
    pass_pct: float = 1.0

    def _update(self):
        if not self.sub_tests:
            self.pass_pct = 1.0 if self.status == Status.PASS else 0
        else:
            self.pass_pct = 1.0
            for sub_test in self.sub_tests:
                if not sub_test.is_passing():
                    self.pass_pct = 0.0
                    break

            # No way to get all subtests, only the ones that failed. So this is
            # if fruitless. Just 1 or 0 is all we get. I.e., avoid subtests.

            # tot = 0
            # for test in self.sub_tests:
            #     tot += (1.0 if test.status == Status.PASS else 0)
            # self.pass_pct = tot / len(self.sub_tests)
        self.status = Status.PASS if self.pass_pct == 1 else Status.FAIL

    def get_subtest(self, name: str) -> SubTest:
        subtest = SubTest(name, self.name, self.parent_test_class)
        self.sub_tests.append(subtest)
        return subtest

    def get_pass_pct(self) -> float:
        self._update()
        return self.pass_pct

    def fail(self, message: str = None) -> None:
        """Indicate this test failed."""
        self.status = Status.FAIL
        self.message = message
        self._update()

    def error(self, message: str = None) -> None:
        """Indicate this test encountered an error."""
        self.status = Status.ERROR
        self.message = message
        self._update()

    def is_passing(self):
        """Check if the test is currently passing."""
        return self.status is Status.PASS


@dataclass
class TestClass:
    name: str
    status: Status = Status.PASS
    tests: Dict[str, Test] = field(default_factory=dict)
    pass_pct: float = 1.0
    pass_count: int = 0

    def _update(self):
        self.pass_count = 0
        tot = 0
        for test in self.tests.values():
            tot += test.get_pass_pct()
            if test.is_passing():
                self.pass_count += 1
        self.pass_pct = tot / len(self.tests)
        self.status = Status.PASS if self.pass_pct == 1 else Status.FAIL

    def get_test(self, name: str):
        """Retrieve the test object"""
        return self.tests.setdefault(name, Test(name, self.name))

    def is_passing(self):
        """Check if the test is currently passing."""
        self._update()
        return self.status is Status.PASS

    def get_pass_pct(self) -> float:
        self._update()
        return self.pass_pct

@dataclass
class Results:
    """Overall results of a test run."""

    status: Status = Status.PASS
    message: str|None = None
    test_classes: Dict[str, TestClass] = field(default_factory=dict)
    pass_pct: float = 1.0

    def get_test(self, nodeid: str):
        """Create or retrieve a Test instance for a given test."""
        node = NodeId.parse(nodeid)
        # Create or receive the parent TestClass
        test_class = self.test_classes.setdefault(node.test_class, TestClass(node.test_class))
        return test_class.get_test(node.test_name)

    def update(self) -> None:
        if not self.test_classes:
            self.pass_pct = 1.0 if self.status == Status.PASS else 0
        else:
            tot = 0
            for test_class in self.test_classes.values():
                tot += test_class.get_pass_pct()
            self.pass_pct = tot / len(self.test_classes)
        self.status = Status.PASS if self.pass_pct == 1 else Status.FAIL

    def error(self, message: str = None) -> None:
        """Indicate the test run fatally errored."""
        self.status = Status.ERROR
        self.message = message
        self.update()

    @staticmethod
    def _dict_factory(items):
        result = {}
        for key, value in items:
            if isinstance(value, Status):
                value = value.name.lower()

            result[key] = value
        
        return result
    
    def as_dict(self):
        self.update()
        return asdict(self, dict_factory=self._dict_factory)

    def as_json(self):
        results = self.as_dict()
        return json.dumps(results, indent=2)
    
    @staticmethod
    def from_dict(results_d: Dict[str, any]) -> 'Results':
        return from_dict(Results, results_d, config=Config(cast=[Enum]))
    


from enum import Enum, auto
from typing import Dict, overload

from pathlib import Path
from pprint import pformat

import logging

logger = logging.getLogger(__name__)


class FileType(Enum):
    """
    Enum for FILE, DIRECTORY or BOTH
    """

    FILE = auto()
    DIRECTORY = auto()
    BOTH = auto()


@overload
def label_dir(
    pattern_dict: Dict[str, str],
    search_dir: Path,
    file_type: FileType = FileType.FILE,
    ignore_case: bool = True,
    skip_hidden: bool = True,
) -> Dict[str, str]: ...


@overload
def label_dir(
    pattern_dict: Dict[str, str],
    search_dir: str,
    file_type: FileType = FileType.FILE,
    ignore_case: bool = True,
    skip_hidden: bool = True,
) -> Dict[str, str]: ...


def label_dir(
    pattern_dict: Dict[str, str],
    search_dir: Path | str,
    file_type: FileType = FileType.FILE,
    ignore_case: bool = True,
    skip_hidden: bool = True,
) -> Dict[str, str]:
    """
    Given a `pattern_dict` containing search patterns as keys and labels as
    values, return a dictionary of all files or subdirectories (sub-objects) of
    `search_dir` with an associated label. A label `pattern_dict` is applied to
    a sub-object if the its name contains the corresponding pattern. If no
    pattern is not found in the dir, then the associated label is set to None.

    For example, assume pattern_dict of:
        {
            'foo': 'fat', 'bar': 'bat'
        }

    Also assume root_dir contains 3 entries:
        giga_foo    kilo_bar    mega_moo

    This function will return:

        {
            'giga_foo': 'fat', 'baby_bar': 'bat', 'mega_moo': None
        }
    """

    if isinstance(search_dir, str):
        root_dir = Path(search_dir).resolve()
    elif isinstance(search_dir, Path):
        root_dir = search_dir.resolve()
    else:
        raise ValueError("root_dir must be a str or a Path")

    if not root_dir.is_dir():
        raise FileNotFoundError(f"{root_dir} is not a valid directory")

    obj_label_map = {}

    logger.debug(f"Checking {root_dir} for subdirs")
    for obj in root_dir.iterdir():
        if obj.is_dir() and file_type == FileType.FILE:
            continue
        if obj.is_file() and file_type == FileType.DIRECTORY:
            continue
        if skip_hidden and obj.name.startswith("."):
            continue
        obj_label_map[obj.name] = None

    logger.debug(f"found {len(obj_label_map)} subdirs")

    # Sort the list of patterns largest to smallest to avoid potential issues
    # with patterns being substrings of one another

    sorted_patterns = list(pattern_dict.keys())
    sorted_patterns.sort(key=lambda s: len(s), reverse=True)
    logger.debug(f"Ordered substitution patterns:\n{pformat(sorted_patterns)}")

    for obj_name in obj_label_map:
        match_obj_name = obj_name.lower() if ignore_case else obj_name
        for pattern in sorted_patterns:
            match_pattern = pattern.lower() if ignore_case else pattern
            if match_pattern in match_obj_name:
                label = pattern_dict[pattern]
                obj_label_map[obj_name] = label
                break

    return obj_label_map



import sys
from pathlib import Path
import logging

import click
from rich.console import Console

from f_table import BasicScreenStyle, get_table
from git_wrap import get_git_repos

from au.click import DebugOptions, BasePath


logger = logging.getLogger(__name__)


@click.command()
@click.argument("root_dir", type=BasePath(resolve_path=True), default=".")
@click.option(
    "-m",
    "--message",
    prompt=True,
    type=str,
    default="Posting feedback",
    help="the message to apply to all `git commit` calls",
)
@click.option(
    "-p",
    "--preview",
    is_flag=True,
    help="set to show changes without actually making them",
)
@DebugOptions().options
def commit_all(
    root_dir: Path,
    message: str = "Posting feedback",
    preview: bool = False,
    quiet: bool = False,
    debug: bool = False,
):
    """Commit and push all "dirty" student repos in ROOT_DIR.

    Iterate over all immediate subdirectories of ROOT_DIR. If it is found to be
    a Git repository and if it contains changes, then:

    \b
        + add all changes
        + commit using --message
        + push all changes to the remote

    If ROOT_DIR is not provided, then the current working directory will be
    assumed.

    If the `--message` argument is not provided, the script will prompt for one.
    """
    logging.basicConfig()

    skips = []
    commits = []
    errors = []

    console = Console()
    with console.status(
        "Finding all local Git repositories", spinner="bouncingBall"
    ) as status:

        all_repos = get_git_repos(root_dir)

        for repo in all_repos:
            status.update(status=f"{repo.name}: Checking for changes")
            if repo.name.startswith("_"):
                skips.append([repo.name, "special directory"])
                continue

            if not repo.is_dirty():
                skips.append([repo.name, "no changes to commit"])
                continue

            if preview:
                commits.append([repo.name, "WOULD COMMIT"])
                continue

            try:
                status.update(status=f"{repo.name}: git pull")
                repo.pull()

                status.update(status=f"{repo.name}: git add .")
                repo.add()

                status.update(status=f"{repo.name}: git commit -m {message}")
                repo.commit(message)

                status.update(status=f"{repo.name}: git push")
                repo.push()

                commits.append([repo.name, "COMMITTED"])
            except Exception as ex:
                logger.exception("Error occurred running git command")
                errors.append([repo.name, str(ex)])

    # Print a summary
    if quiet:
        all_dirs = errors
    else:
        all_dirs = commits + skips + errors
    all_dirs.sort()

    if not all_dirs:
        print("No repositories found")
        sys.exit(0)

    print(get_table(all_dirs, col_defs=["", "A"], style=BasicScreenStyle()))

    summary = []
    summary.append([f"Root Directory", root_dir])
    if preview:
        summary.append([f"Repositories to Commit", len(commits)])
    else:
        summary.append([f"Repositories Pushed", len(commits)])
    summary.append([f"Directories Skipped", len(skips)])
    summary.append([f"Errors Encountered", len(errors)])
    print(get_table(summary, style=BasicScreenStyle()))


if __name__ == "__main__":
    commit_all()



from textwrap import fill, dedent
from os.path import dirname, realpath

# print a list of items with word wrapping
def ps(lines):
    for line in lines:
        print(fill(dedent(line).strip(), 80))

# print a list of items as a wrapped bulleted list
def pb(lines):
    for line in lines:
        print(fill(line.strip(), 80, initial_indent="  • ", subsequent_indent="   "))

# write a list of items as a wrapped bulleted list to a file
def wb(lines, fi):
    for line in lines:
        fi.write(fill(line.strip(), 80, initial_indent="  * ", subsequent_indent="    "))
        fi.write('\n')

# retrieve the fill path of a file in the relative to the real path of current script
def fullpath(filename):
    path = dirname(realpath(__file__))
    return path + '/' + filename
